from datetime import datetime, date
from typing import Optional
from uuid import UUID

from .models import ValidateBaseModel as BaseModel

from ..enums import (
    ActivityType,
    NonTradeActivityStatus,
    OrderStatus,
    OrderSide,
    TradeActivityType,
    AccountStatus,
)


class BaseActivity(BaseModel):
    """
    Represents Base information for an event/activity for a specific Account.

    You most likely will want an instance of one of the child classes TradeActivity and NonTradeActivity

    Attributes:
        id (str): Unique ID of this Activity. Note that IDs for Activity instances are formatted like
          `20220203000000000::045b3b8d-c566-4bef-b741-2bf598dd6ae7` the first part before the `::` is a date string
          while the part after is a UUID
        account_id (UUID): id of the Account this activity relates too
        activity_type (ActivityType): What specific kind of Activity this was
    """

    id: str
    account_id: UUID
    activity_type: ActivityType

    def __init__(self, *args, **data):
        if "account_id" in data and type(data["account_id"]) == str:
            data["account_id"] = UUID(data["account_id"])

        super().__init__(*args, **data)


class NonTradeActivity(BaseActivity):
    """
    A NonTradeActivity represents an Activity that happened for an account that doesn't have to do with orders or trades.

    Attributes:
        date (date): The date on which the activity occurred or on which the transaction associated with the
          activity settled.
        net_amount (float): The net amount of money (positive or negative) associated with the activity.
        description (str): Extra description of the NTA if needed. Can be empty string ""
        status (NonTradeActivityStatus): Status of the activity. Not present for all activity types.
        symbol (Optional[str]): The symbol of the security involved with the activity. Not present for all activity
          types.
        qty (Optional[float]): For dividend activities, the number of shares that contributed to the payment. Not
          present for other activity types.
        price (Optional[float]) Not present for all activity types.
        per_share_amount (Optional[float]): For dividend activities, the average amount paid per share. Not present for
          other activity types.
    """

    date: date
    net_amount: float
    description: str
    status: Optional[NonTradeActivityStatus] = None
    symbol: Optional[str] = None
    qty: Optional[float] = None
    price: Optional[float] = None
    per_share_amount: Optional[float] = None


class TradeActivity(BaseActivity):
    """
    Represents information for TradeActivities. TradeActivities are Activities that pertain to trades that happened for
    an account. IE Fills or partial fills for orders.

    Attributes:
        transaction_time (datetime): The time and date of when this trade was processed
        type (TradeActivityType): What kind of trade this TradeActivity represents. See TradeActivityType for more
          details
        price (float): The per-share price that the trade was executed at.
        qty (float): The number of shares involved in the trade execution.
        side (OrderSide): What side the trade this TradeActivity represents was on. See OrderSide for more information
        symbol (str): The symbol of the asset that was traded
        leaves_qty (float): For partially filled orders, the quantity of shares that are left to be filled. Will be 0 if
          order was not a partially filled order
        order_id (UUID): The ID for the order filled
        cum_qty (float): The cumulative quantity of shares involved in the execution.
        order_status (OrderStatus): The status of the order that executed the trade. See OrderStatus for more details
    """

    transaction_time: datetime
    type: TradeActivityType
    price: float
    qty: float
    side: OrderSide
    symbol: str
    leaves_qty: float
    order_id: UUID
    cum_qty: float
    order_status: OrderStatus


class TradeAccount(BaseModel):
    """
    Represents trading account information for an Account.

    Attributes:
        id (UUID): The account ID
        account_number (str): The account number
        status (AccountStatus): The current status of the account
        crypto_status (Optional[AccountStatus]): The status of the account in regards to trading crypto. Only present if
          crypto trading is enabled for your brokerage account.
        currency (Optional[str]): Currently will always be the value "USD".
        buying_power (str): Current available cash buying power. If multiplier = 2 then
          buying_power = max(equity-initial_margin(0) * 2). If multiplier = 1 then buying_power = cash.
        regt_buying_power (str): User’s buying power under Regulation T
          (excess equity - (equity - margin value) - * margin multiplier)
        daytrading_buying_power (str): The buying power for day trades for the account
        non_marginable_buying_power (str): The non marginable buying power for the account
        cash (str): Cash balance in the account
        accrued_fees (str): Fees accrued in this account
        pending_transfer_out (Optional[str]): Cash pending transfer out of this account
        pending_transfer_in (Optional[str]): Cash pending transfer into this account
        portfolio_value (str): Total value of cash + holding positions.
          (This field is deprecated. It is equivalent to the equity field.)
        pattern_day_trader (bool): Whether the account is flagged as pattern day trader or not.
        trading_blocked (bool): If true, the account is not allowed to place orders.
        transfers_blocked (bool): If true, the account is not allowed to request money transfers.
        account_blocked (bool): If true, the account activity by user is prohibited.
        created_at (datetime): Timestamp this account was created at
        trade_suspended_by_user (bool): If true, the account is not allowed to place orders.
        multiplier (str): Multiplier value for this account.
        shorting_enabled (bool): Flag to denote whether or not the account is permitted to short
        equity (str): This value is cash + long_market_value + short_market_value. This value isn't calculated in the
          SDK it is computed on the server and we return the raw value here.
        last_equity (str): Equity as of previous trading day at 16:00:00 ET
        long_market_value (str): Real-time MtM value of all long positions held in the account
        short_market_value (str): Real-time MtM value of all short positions held in the account
        initial_margin (str): Reg T initial margin requirement
        maintenance_margin (str): Maintenance margin requirement
        last_maintenance_margin (str): Maintenance margin requirement on the previous trading day
        sma (str): Value of Special Memorandum Account (will be used at a later date to provide additional buying_power)
        daytrade_count (int): The current number of daytrades that have been made in the last 5 trading days
          (inclusive of today)
    """

    id: UUID
    account_number: str
    status: AccountStatus
    crypto_status: Optional[AccountStatus]
    currency: Optional[str]
    buying_power: str
    regt_buying_power: str
    daytrading_buying_power: str
    non_marginable_buying_power: str
    cash: str
    accrued_fees: str
    pending_transfer_out: Optional[str]
    pending_transfer_in: Optional[str]
    portfolio_value: str
    pattern_day_trader: bool
    trading_blocked: bool
    transfers_blocked: bool
    account_blocked: bool
    created_at: datetime
    trade_suspended_by_user: bool
    multiplier: str
    shorting_enabled: bool
    equity: str
    last_equity: str
    long_market_value: str
    short_market_value: str
    initial_margin: str
    maintenance_margin: str
    last_maintenance_margin: str
    sma: str
    daytrade_count: int



from uuid import UUID
from .models import ValidateBaseModel as BaseModel
from datetime import datetime
from .trading import Asset
from typing import List, Optional


class Watchlist(BaseModel):
    """
    A watchlist is an ordered list of assets. An account can have multiple watchlists.
    Learn more about watchlists in the documentation. https://alpaca.markets/docs/api-references/trading-api/watchlist/

    Attributes:
        account_id (UUID): The uuid identifying the account the watchlist belongs to
        id (UUID): The unique identifier for the watchlist
        name (str): An arbitrary string up to 64 characters identifying the watchlist
        created_at (datetime): When the watchlist was created
        updated_at (datetime): When the watchlist was last updated
        assets (Optional[List[Asset]]): The assets in the watchlist, not returned from all endpoints
    """

    account_id: UUID
    id: UUID
    name: str
    created_at: datetime
    updated_at: datetime
    assets: Optional[List[Asset]]



from datetime import date
from typing import Optional, List
from uuid import UUID

from pydantic import root_validator

from alpaca.broker.enums import JournalEntryType, JournalStatus
from alpaca.common.models import ValidateBaseModel as BaseModel, NonEmptyRequest


class Journal(BaseModel):
    """
    Represents a transfer of cash or securities from one account to another.

    There are two types of journals Cash Journals and Security Journals.

    **Travel Rule**
    In an effort to fight the criminal financial transactions, FinCEN enacted the
    Travel Rule that applies to fund transfers of more than $3,000.
    When you use Journal API to bundle a bulk of transfers for the end-users, you will need to tell about
    the breakdown and each transmitter information using the optional fields of the POST request.

    Learn more about journals here: https://alpaca.markets/docs/api-references/broker-api/journals/

    Attributes:
        id (UUID): The journal ID
        to_account (UUID): The account ID that received the journal.
        from_account (UUID): The account ID that initiates the journal.
        entry_type (JournalEntryType): Whether the journal is a cash or security journal.
        status (JournalStatus): The lifecycle status of the journal.
        symbol (Optional[str]): For security journals, the symbol identifier of the security being journaled.
        qty (Optional[float]): For security journals, the quantity of the security being journaled.
        price (Optional[float]): For security journals, the price at which the security is being journaled at.
        net_amount (Optional[float]): For cash journals, the total cash amount journaled
        description (Optional[str]): Journal description. It can include fixtures for sandbox API.
        settle_date (Optional[date]):
        system_date (Optional[date]):
        transmitter_name (Optional[str]): For cash journals, travel rule related name info.
        transmitter_account_number (Optional[str]): For cash journals, travel rule account number info.
        transmitter_address (Optional[str]): For cash journals, travel rule related address info.
        transmitter_financial_institution (Optional[str]): For cash journals, travel rule related institution info.
        transmitter_timestamp (Optional[str]): For cash journals, travel rule related timestamp info.
    """

    id: UUID
    to_account: UUID
    from_account: UUID
    entry_type: JournalEntryType
    status: JournalStatus
    net_amount: Optional[float]
    symbol: Optional[str]
    qty: Optional[float]
    price: Optional[float]
    description: Optional[str]
    settle_date: Optional[date]
    system_date: Optional[date]
    transmitter_name: Optional[str]
    transmitter_account_number: Optional[str]
    transmitter_address: Optional[str]
    transmitter_financial_institution: Optional[str]
    transmitter_timestamp: Optional[str]


class BatchJournalResponse(Journal):
    """
    Represents a journal response from a batch journal request.

    Attributes:
        error_message (Optional[str]): An message that contains error details for failed journals.
    """

    error_message: Optional[str]


class CreateJournalRequest(NonEmptyRequest):
    """
    Data for request to initiate a single journal.

    Attributes:
        to_account (UUID): The account ID that received the journal.
        from_account (UUID): The account ID that initiates the journal.
        entry_type (JournalEntryType): Whether the journal is a cash or security journal.
        symbol (Optional[str]): For security journals, the symbol identifier of the security being journaled.
        qty (Optional[float]): For security journals, the quantity of the security being journaled.
        amount (Optional[float]): For cash journals, the total cash amount journaled in USD.
        description (Optional[str]): Journal description. It can include fixtures for sandbox API.
        transmitter_name (Optional[str]): For cash journals, travel rule related name info.
        transmitter_account_number (Optional[str]): For cash journals, travel rule account number info.
        transmitter_address (Optional[str]): For cash journals, travel rule related address info.
        transmitter_financial_institution (Optional[str]): For cash journals, travel rule related institution info.
        transmitter_timestamp (Optional[str]): For cash journals, travel rule related timestamp info.
    """

    from_account: UUID
    entry_type: JournalEntryType
    to_account: UUID
    amount: Optional[float]
    symbol: Optional[str]
    qty: Optional[float]
    description: Optional[str]
    transmitter_name: Optional[str]
    transmitter_account_number: Optional[str]
    transmitter_address: Optional[str]
    transmitter_financial_institution: Optional[str]
    transmitter_timestamp: Optional[str]

    @root_validator()
    def root_validator(cls, values: dict) -> dict:

        entry_type = values.get("entry_type")
        symbol = values.get("symbol")
        qty = values.get("qty")
        amount = values.get("amount")

        # amount is for cash journals, symbol and qty are for security journals
        # they are mutually exclusive
        if entry_type is not None and entry_type == JournalEntryType.CASH:
            if symbol or qty:
                raise ValueError("Symbol and qty are reserved for security journals.")

            if not amount:
                raise ValueError("Cash journals must contain an amount to transfer.")

        if entry_type is not None and entry_type == JournalEntryType.SECURITY:

            if amount:
                raise ValueError("Amount is reserved for cash journals.")

            if not symbol or not qty:
                raise ValueError(
                    "Security journals must contain a symbol and corresponding qty to transfer."
                )

        return values


class BatchJournalRequestEntry(NonEmptyRequest):
    """
    Entry in batch journal request.

    Attributes:
        to_account (UUID): Account to fund in batch journal request.
        amount (float): The cash amount in USD to fund by.
    """

    to_account: UUID
    amount: float


class ReverseBatchJournalRequestEntry(NonEmptyRequest):
    """
    Entry in reverse batch journal request.

    Attributes:
        to_account (UUID): Account to fund in batch journal request.
        amount (float): The cash amount in USD to fund by.
    """

    from_account: UUID
    amount: float


class CreateBatchJournalRequest(NonEmptyRequest):
    """
    This model represents the fields you can specify when creating
    a request of many Journals out of one account to many others at once.

    Currently, batch journals are only enabled on cash journals.

    Attributes:
        entry_type (JournalEntryType): The type of journal transfer.
        from_account (UUID): The originator of funds. Most likely is your Sweep Firm Account
        description (Optional[str]): Journal description, gets returned in the response.
        entries (List[BatchJournalRequestEntry]): List of journals to execute.
    """

    entry_type: JournalEntryType
    from_account: UUID
    description: Optional[str]
    entries: List[BatchJournalRequestEntry]


class CreateReverseBatchJournalRequest(NonEmptyRequest):
    """
    This model represents the fields you can specify when creating
    a request of many Journals into one account from many other accounts at once.

    Currently, reverse batch journals are only enabled on cash journals.

    Attributes:
        entry_type (JournalEntryType): The type of journal transfer.
        to_account (UUID): The destination of funds. Most likely is your Sweep Firm Account
        description (Optional[str]): Journal description, gets returned in the response.
        entries (List[BatchJournalRequestEntry]): List of journals to execute.
    """

    entry_type: JournalEntryType
    to_account: UUID
    description: Optional[str]
    entries: List[ReverseBatchJournalRequestEntry]


class GetJournalsRequest(NonEmptyRequest):
    """
    This model represents the fields you can specify when querying from the list of all journals.

    Attributes:
        after (Optional[date]): Journal creation dates after this date.
        before (Optional[date]): Journal creation dates before this date.
        status (Optional[JournalStatus]): Only journals with this status.
        entry_type (Optional[JournalEntryType]): Only journals with this entry type.
        to_account (Optional[UUID]): Only journals to this account.
        from_account (Optional[UUID]): Only journals from this account.
    """

    after: Optional[date]
    before: Optional[date]
    status: Optional[JournalStatus]
    entry_type: Optional[JournalEntryType]
    to_account: Optional[UUID]
    from_account: Optional[UUID]



from datetime import date
from uuid import UUID

from .models import ValidateBaseModel as BaseModel
from ..enums import CorporateActionType, CorporateActionSubType


class CorporateActionAnnouncement(BaseModel):
    """
    An announcement of a corporate action. Corporate actions are events like dividend payouts, mergers and stock splits.

    Attributes:
        id (UUID): The unique identifier for this single announcement.
        corporate_action_id (str): ID that remains consistent across all announcements for the same corporate action.
        ca_type (CorporateActionType): The type of corporate action that was announced.
        ca_sub_type (CorporateActionSubType): The specific subtype of corporate action that was announced.
        initiating_symbol (str): Symbol of the company initiating the announcement.
        initiating_original_cusip (str): CUSIP of the company initiating the announcement.
        target_symbol (str): Symbol of the child company involved in the announcement.
        target_original_cusip (str): CUSIP of the child company involved in the announcement.
        declaration_date (date): Date the corporate action or subsequent terms update was announced.
        ex_date (date): The first date that purchasing a security will not result in a corporate action entitlement.
        record_date (date): The date an account must hold a settled position in the security in order to receive the
            corporate action entitlement.
        payable_date (date): The date the announcement will take effect. On this date, account stock and cash
            balances are expected to be processed accordingly.
        cash (float): The amount of cash to be paid per share held by an account on the record date.
        old_rate (float): The denominator to determine any quantity change ratios in positions.
        new_rate (float): The numerator to determine any quantity change ratios in positions.
    """

    id: UUID
    corporate_action_id: str
    ca_type: CorporateActionType
    ca_sub_type: CorporateActionSubType
    initiating_symbol: str
    initiating_original_cusip: str
    target_symbol: str
    target_original_cusip: str
    declaration_date: date
    ex_date: date
    record_date: date
    payable_date: date
    cash: float
    old_rate: float
    new_rate: float



from typing import Union, List, Optional

from alpaca.common import NonEmptyRequest
from alpaca.data import DataFeed


class StockSnapshotRequest(NonEmptyRequest):
    """
    This request class is used to submit a request for snapshot data for stocks.

    Attributes:
        symbol_or_symbols (Union[str, List[str]]): The ticker identifier or list of ticker identifiers.
        feed (Optional[DataFeed]): The stock data feed to retrieve from.
    """

    symbol_or_symbols: Union[str, List[str]]
    feed: Optional[DataFeed]


class CryptoSnapshotRequest(NonEmptyRequest):
    """
    This request class is used to submit a request for snapshot data for crypto.

    Attributes:
        symbol_or_symbols (Union[str, List[str]]): The ticker identifier or list of ticker identifiers.
    """

    symbol_or_symbols: Union[str, List[str]]



from alpaca.common.websocket import BaseStream
from typing import Optional, Dict
from alpaca.common.enums import BaseURL


class CryptoDataStream(BaseStream):
    """
    A WebSocket client for streaming live crypto data.

    See BaseStream for more information on implementation and the methods available.
    """

    def __init__(
        self,
        api_key: str = None,
        secret_key: str = None,
        raw_data: bool = False,
        websocket_params: Optional[Dict] = None,
    ) -> None:
        """
        Instantiates a WebSocket client for accessing live cryptocurrency data.

        Args:
            api_key (str): Alpaca API key. Defaults to None.
            secret_key (str): Alpaca API secret key. Defaults to None.
            raw_data (bool, optional): Whether to return wrapped data or raw API data. Defaults to False.
            websocket_params (Optional[Dict], optional): Any parameters for configuring websocket connection. Defaults to None.
        """
        super().__init__(
            endpoint=BaseURL.MARKET_DATA_LIVE.value + "/v1beta1/crypto",
            api_key=api_key,
            secret_key=secret_key,
            raw_data=raw_data,
            websocket_params=websocket_params,
        )



from collections import defaultdict
from typing import Union, Optional, List

from alpaca.common import RawData, DATA_V2_MAX_LIMIT, BaseURL
from alpaca.common.rest import RESTClient, HTTPResult
from alpaca.data import BarSet, QuoteSet, TradeSet, SnapshotSet
from alpaca.data.historical.stock import DataExtensionType
from alpaca.data.requests import (
    CryptoBarsRequest,
    CryptoQuotesRequest,
    CryptoTradesRequest,
    LatestCryptoTradeRequest,
    LatestCryptoQuoteRequest,
    CryptoSnapshotRequest,
)


class CryptoHistoricalDataClient(RESTClient):
    def __init__(
        self,
        api_key: Optional[str] = None,
        secret_key: Optional[str] = None,
        oauth_token: Optional[str] = None,
        raw_data: bool = False,
        url_override: Optional[str] = None,
    ) -> None:
        """
        Instantiates a Historical Data Client for Crypto Data.

        Args:
            api_key (Optional[str], optional): Alpaca API key. Defaults to None.
            secret_key (Optional[str], optional): Alpaca API secret key. Defaults to None.
            oauth_token (Optional[str]): The oauth token if authenticating via OAuth. Defaults to None.
            raw_data (bool, optional): If true, API responses will not be wrapped and raw responses will be returned from
              methods. Defaults to False.
            url_override (Optional[str], optional): If specified allows you to override the base url the client points
              to for proxy/testing.
        """
        super().__init__(
            api_key=api_key,
            secret_key=secret_key,
            oauth_token=oauth_token,
            api_version="v2",
            base_url=url_override if url_override is not None else BaseURL.DATA,
            sandbox=False,
            raw_data=raw_data,
        )

    def get_crypto_bars(
        self, request_params: CryptoBarsRequest
    ) -> Union[BarSet, RawData]:
        """Gets bar/candle data for a cryptocurrency or list of cryptocurrencies.

        Args:


        Returns:
            Union[BarSet, RawData]: The crypto bar data either in raw or wrapped form
        """

        params = request_params.to_request_fields()

        # paginated get request for crypto market data api
        raw_bars = self._data_get(
            endpoint_asset_class="crypto",
            endpoint_data_type="bars",
            api_version="v1beta2",
            **params,
        )

        return self.response_wrapper(model=BarSet, raw_data=raw_bars)

    def get_crypto_quotes(
        self, request_params: CryptoQuotesRequest
    ) -> Union[QuoteSet, RawData]:
        """Returns Quote level 1 data over a given time period for a cryptocurrency or list of cryptocurrencies.

        Args:


        Returns:
            Union[QuoteSet, RawData]: The quote data either in raw or wrapped form
        """

        params = request_params.to_request_fields()

        # paginated get request for market data api
        raw_quotes = self._data_get(
            endpoint_asset_class="crypto",
            endpoint_data_type="quotes",
            api_version="v1beta2",
            **params,
        )

        return self.response_wrapper(model=QuoteSet, raw_data=raw_quotes)

    def get_crypto_trades(
        self, request_params: CryptoTradesRequest
    ) -> Union[TradeSet, RawData]:
        """Returns the price and sales history over a given time period for a cryptocurrency or list of cryptocurrencies.

        Args:


        Returns:
            Union[TradeSet, RawData]: The trade data either in raw or wrapped form
        """

        params = request_params.to_request_fields()

        # paginated get request for market data api
        raw_trades = self._data_get(
            endpoint_asset_class="crypto",
            endpoint_data_type="trades",
            api_version="v1beta2",
            **params,
        )

        return self.response_wrapper(model=TradeSet, raw_data=raw_trades)

    def get_crypto_latest_trade(
        self, request_params: LatestCryptoTradeRequest
    ) -> Union[TradeSet, RawData]:
        """Returns the latest trade for a coin for a specific exchange

        Args:


        Returns:
            Union[Trade, RawData]: The latest trade in raw or wrapped format
        """

        params = request_params.to_request_fields()

        raw_trades = self._data_get(
            endpoint_asset_class="crypto",
            endpoint_data_type="trades",
            api_version="v1beta2",
            extension=DataExtensionType.LATEST,
            **params,
        )

        return self.response_wrapper(model=TradeSet, raw_data=raw_trades)

    def get_crypto_latest_quote(
        self, request_params: LatestCryptoQuoteRequest
    ) -> Union[QuoteSet, RawData]:
        """Returns the latest quote for a coin for a specific exchange

                Args:
        =

                Returns:
                    Union[Quote, RawData]: The latest quote in raw or wrapped format
        """

        params = request_params.to_request_fields()

        raw_quotes = self._data_get(
            endpoint_asset_class="crypto",
            endpoint_data_type="quotes",
            api_version="v1beta2",
            extension=DataExtensionType.LATEST,
            **params,
        )

        return self.response_wrapper(model=QuoteSet, raw_data=raw_quotes)

    def get_crypto_snapshot(
        self, request_params: CryptoSnapshotRequest
    ) -> Union[SnapshotSet, RawData]:
        """Returns snapshots of queried crypto symbols. Snapshots contain latest trade, latest quote, latest minute bar,
        latest daily bar and previous daily bar data for the queried symbols.

        Args:


        Returns:
            Union[SnapshotSet, RawData]: The snapshot data either in raw or wrapped form
        """

        params = request_params.to_request_fields()

        raw_snapshots = self._data_get(
            endpoint_asset_class="crypto",
            endpoint_data_type="snapshots",
            api_version="v1beta2",
            extension=DataExtensionType.SNAPSHOT,
            **params,
        )
        print(raw_snapshots)
        return self.response_wrapper(
            model=SnapshotSet,
            raw_data=raw_snapshots,
        )

    def _data_get(
        self,
        endpoint_asset_class: str,
        endpoint_data_type: str,
        api_version: str,
        symbol_or_symbols: Union[str, List[str]],
        limit: Optional[int] = None,
        page_limit: int = DATA_V2_MAX_LIMIT,
        extension: Optional[DataExtensionType] = None,
        **kwargs,
    ) -> RawData:
        """Performs Data API GET requests accounting for pagination. Data in responses are limited to the page_limit,
        which defaults to 10,000 items. If any more data is requested, the data will be paginated.

        Args:
            endpoint_data_type (str): The data API endpoint path - /bars, /quotes, etc
            symbol_or_symbols (Union[str, List[str]]): The symbol or list of symbols that we want to query for
            endpoint_asset_class (str): The data API security type path. Defaults to 'stocks'.
            api_version (str): Data API version. Defaults to "v2".
            limit (Optional[int]): The maximum number of items to query. Defaults to None.
            page_limit (Optional[int]): The maximum number of items returned per page - different from limit. Defaults to DATA_V2_MAX_LIMIT.

        Returns:
            RawData: Raw Market data from API
        """
        # params contains the payload data
        params = kwargs

        # stocks, crypto, etc
        path = f"/{endpoint_asset_class}"

        if isinstance(symbol_or_symbols, str):
            symbol_or_symbols = [symbol_or_symbols]

        params["symbols"] = ",".join(symbol_or_symbols)

        # TODO: Improve this mess if possible
        if extension == DataExtensionType.LATEST:
            path += "/latest"
            path += f"/{endpoint_data_type}"
        elif extension == DataExtensionType.SNAPSHOT:
            path += "/snapshots"
        else:
            # bars, trades, quotes, etc
            path += f"/{endpoint_data_type}"

        # data_by_symbol is in format of
        #    {
        #       "symbol1": [ "data1", "data2", ... ],
        #       "symbol2": [ "data1", "data2", ... ],
        #                ....
        #    }
        data_by_symbol = defaultdict(list)

        total_items = 0
        page_token = None

        while True:

            actual_limit = None

            # adjusts the limit parameter value if it is over the page_limit
            if limit:
                # actual_limit is the adjusted total number of items to query per request
                actual_limit = min(int(limit) - total_items, page_limit)
                if actual_limit < 1:
                    break

            params["limit"] = actual_limit
            params["page_token"] = page_token

            response = self.get(path=path, data=params, api_version=api_version)

            # TODO: Merge parsing if possible
            if extension == DataExtensionType.SNAPSHOT:
                self._parse_snapshot(response, data_by_symbol)
            else:
                self._parse_response(response, data_by_symbol)

            # if we've sent a request with a limit, increment count
            if actual_limit:
                total_items += actual_limit

            page_token = response.get("next_page_token", None)

            if page_token is None:
                break

        return data_by_symbol

    @staticmethod
    def _parse_response(response: HTTPResult, data_by_symbol: dict) -> RawData:

        # data_by_symbol is in format of
        #    {
        #       "symbol1": [ "data1", "data2", ... ],
        #       "symbol2": [ "data1", "data2", ... ],
        #                ....
        #    }

        response_data = CryptoHistoricalDataClient.get_data_from_response(response)

        # add elements to data_by_symbol
        # for list data types just extend
        # for non-list types, add as element of a list.
        # list comprehension used for speed
        [
            data_by_symbol[symbol].extend(data)
            if isinstance(data, list)
            else data_by_symbol[symbol].append(data)
            for symbol, data in response_data.items()
        ]

        return data_by_symbol

    @staticmethod
    def get_data_from_response(response: HTTPResult) -> RawData:

        data_keys = {
            "trade",
            "trades",
            "quote",
            "quotes",
            "bar",
            "bars",
            "snapshot",
            "snapshots",
        }

        selected_key = data_keys.intersection(response)

        if selected_key is None or len(selected_key) < 1:
            raise ValueError("The data in response does not match any known keys.")

        # assume selected_key only contains 1 value
        selected_key = selected_key.pop()

        # formatting a single symbol response so that this method
        # always returns a symbol keyed data dictionary
        if "symbol" in response:
            return {response["symbol"]: response[selected_key]}

        return response[selected_key]

    @staticmethod
    def _parse_snapshot(response: HTTPResult, data_by_symbol: dict):
        # TODO: Improve snapshot parsing
        response_data = CryptoHistoricalDataClient.get_data_from_response(response)

        for symbol, data in response_data.items():
            data_by_symbol[symbol] = data

        return data_by_symbol



from .models import *
from .accounts import *
from .clock import Clock
from .calendar import Calendar
from .requests import *
from .trading import *
from .watchlist import Watchlist
from .corporate_actions import CorporateActionAnnouncement



from .models import ValidateBaseModel as BaseModel
from uuid import UUID
from datetime import datetime
from typing import Any, Optional, List
from ..enums import (
    AssetClass,
    AssetStatus,
    AssetExchange,
    OrderStatus,
    OrderType,
    OrderClass,
    TimeInForce,
    OrderSide,
    PositionSide,
)
from pydantic import Field


class Asset(BaseModel):
    """
    Represents a security. Some Assets are not tradable with Alpaca. These Assets are
    marked with the flag `tradable=false`.

    For more info, visit https://alpaca.markets/docs/api-references/trading-api/assets/

    Attributes:
        id (UUID): Unique id of asset
        asset_class (AssetClass): The name of the asset class.
        exchange (AssetExchange): Which exchange this asset is available through.
        symbol (str): The symbol identifier of the asset.
        name (Optional[str]): The name of the asset.
        status (AssetStatus): The active status of the asset.
        tradable (bool): Whether the asset can be traded.
        marginable (bool): Whether the asset can be traded on margin.
        shortable (bool): Whether the asset can be shorted.
        easy_to_borrow (bool): When shorting, whether the asset is easy to borrow
        fractionable (bool): Whether fractional shares are available
    """

    id: UUID
    asset_class: AssetClass = Field(
        alias="class"
    )  # using a pydantic alias to allow parsing data with the `class` keyword field
    exchange: AssetExchange
    symbol: str
    name: Optional[str] = None
    status: AssetStatus
    tradable: bool
    marginable: bool
    shortable: bool
    easy_to_borrow: bool
    fractionable: bool


class Position(BaseModel):
    """
    Represents an open long or short holding in an asset.

    Attributes:
        asset_id (UUID): ID of the asset.
        symbol (str): Symbol of the asset.
        exchange (AssetExchange): Exchange name of the asset.
        asset_class (AssetClass): Name of the asset's asset class.
        avg_entry_price (str): The average entry price of the position.
        qty (str): The number of shares of the position.
        side (PositionSide): "long" or "short" representing the side of the position.
        market_value (str): Total dollar amount of the position.
        cost_basis (str): Total cost basis in dollars.
        unrealized_pl (str): Unrealized profit/loss in dollars.
        unrealized_plpc (str): Unrealized profit/loss percent.
        unrealized_intraday_pl (str): Unrealized profit/loss in dollars for the day.
        unrealized_intraday_plpc (str): Unrealized profit/loss percent for the day.
        current_price (str): Current asset price per share.
        lastday_price (str): Last day’s asset price per share based on the closing value of the last trading day.
        change_today (str): Percent change from last day's price.
    """

    asset_id: UUID
    symbol: str
    exchange: AssetExchange
    asset_class: AssetClass
    avg_entry_price: str
    qty: str
    side: PositionSide
    market_value: str
    cost_basis: str
    unrealized_pl: str
    unrealized_plpc: str
    unrealized_intraday_pl: str
    unrealized_intraday_plpc: str
    current_price: str
    lastday_price: str
    change_today: str


class ClosePositionResponse(BaseModel):
    """
    Attributes:
        order_id (UUID): ID of order that was created to liquidate the position.
        status_code (int): Status code corresponding to the request to liquidate the position.
    """

    order_id: UUID
    status_code: int


class Order(BaseModel):
    """
    Represents a request to buy or sell an asset.

    Attributes:
        id (UUID): Order ID generated by Alpaca.
        client_order_id (str): Client unique order ID
        created_at (datetime): Timestamp when the Order was created.
        updated_at (datetime): Timestamp when the Order was last updated.
        submitted_at (datetime): Timestamp when the Order was submitted.
        filled_at (Optional[datetime]): Timestamp when the Order was filled.
        expired_at (Optional[datetime]): Timestamp when the Order expired at.
        canceled_at (Optional[datetime]): Timestamp when the Order was canceled.
        failed_at (Optional[datetime]): Timestamp when the Order failed at.
        replaced_at (Optional[datetime]): Timestamp when the Order was replaced by a new Order.
        replaced_by (Optional[UUID]): ID of Order that replaces this Order.
        replaces (Optional[UUID]): ID of Order which this Order replaces.
        asset_id (UUID): ID of the asset.
        symbol (str): Symbol of the asset.
        asset_class (AssetClass): Asset class of the asset.
        notional (Optional[str]): Ordered notional amount. If entered, qty will be null. Can take up to 9 decimal
          points.
        qty (Optional[str]): Ordered quantity. If entered, notional will be null. Can take up to 9 decimal points.
        filled_qty (Optional[str]): Filled quantity.
        filled_avg_price (Optional[str]): Filled average price. Can be 0 until order is processed in case order is
          passed outside of market hours.
        order_class (OrderClass): Valid values: simple, bracket, oco or oto.
        order_type (OrderType): Deprecated with just type field below.
        type (OrderType): Valid values: market, limit, stop, stop_limit, trailing_stop.
        side (OrderSide): Valid values: buy and sell.
        time_in_force (TimeInForce): Length of time the Order is in force.
        limit_price (Optional[str]): Limit price of the Order.
        stop_price (Optional[str]): Stop price of the Order.
        status (OrderStatus): The status of the Order.
        extended_hours (bool): If true, eligible for execution outside regular trading hours.
        legs (Optional[List[Order]]): When querying non-simple order_class orders in a nested style, an array of Order
          entities associated with this order. Otherwise, null.
        trail_percent (Optional[str]): The percent value away from the high water mark for trailing stop orders.
        trail_price (Optional[str]): The dollar value away from the high water mark for trailing stop orders.
        hwm (Optional[str]): The highest (lowest) market price seen since the trailing stop order was submitted.
    """

    id: UUID
    client_order_id: str
    created_at: datetime
    updated_at: datetime
    submitted_at: datetime
    filled_at: Optional[datetime]
    expired_at: Optional[datetime]
    canceled_at: Optional[datetime]
    failed_at: Optional[datetime]
    replaced_at: Optional[datetime]
    replaced_by: Optional[UUID]
    replaces: Optional[UUID]
    asset_id: UUID
    symbol: str
    asset_class: AssetClass
    notional: Optional[str]
    qty: Optional[str]
    filled_qty: Optional[str]
    filled_avg_price: Optional[str]
    order_class: OrderClass
    order_type: OrderType
    type: OrderType
    side: OrderSide
    time_in_force: TimeInForce
    limit_price: Optional[str]
    stop_price: Optional[str]
    status: OrderStatus
    extended_hours: bool
    legs: Optional[List["Order"]]
    trail_percent: Optional[str]
    trail_price: Optional[str]
    hwm: Optional[str]

    def __init__(self, **data: Any) -> None:

        if "order_class" not in data or data["order_class"] == "":
            data["order_class"] = OrderClass.SIMPLE

        super().__init__(**data)


class PortfolioHistory(BaseModel):
    """
    Contains information about the value of a portfolio over time.

    Attributes:
        timestamp (List[int]): Time of each data element, left-labeled (the beginning of time window).
        equity (List[float]): Equity value of the account in dollar amount as of the end of each time window.
        profit_loss (List[float]): Profit/loss in dollar from the base value.
        profit_loss_pct (List[float]): Profit/loss in percentage from the base value.
        base_value (float): Basis in dollar of the profit loss calculation.
        timeframe (str): Time window size of each data element.
    """

    timestamp: List[int]
    equity: List[float]
    profit_loss: List[float]
    profit_loss_pct: List[float]
    base_value: float
    timeframe: str



from typing import Union, List, Optional

from alpaca.common import NonEmptyRequest
from alpaca.data import DataFeed

# ############################## Stocks ################################# #


class BaseLatestStockDataRequest(NonEmptyRequest):
    """
    A base request object for retrieving the latest data for stocks. You most likely should not use this directly and
    instead use the asset class specific request objects.

    Attributes:
        symbol_or_symbols (Union[str, List[str]]): The ticker identifier or list of ticker identifiers.
        feed (Optional[DataFeed]): The stock data feed to retrieve from.
    """

    symbol_or_symbols: Union[str, List[str]]
    feed: Optional[DataFeed]


class LatestStockTradeRequest(BaseLatestStockDataRequest):
    """
    This request class is used to submit a request for the latest stock trade data.

    See BaseLatestStockDataRequest for more information on available parameters.
    """

    pass


class LatestStockQuoteRequest(BaseLatestStockDataRequest):
    """
    This request class is used to submit a request for the latest stock quote data.

    See BaseLatestStockDataRequest for more information on available parameters.
    """

    pass


class LatestStockBarRequest(BaseLatestStockDataRequest):
    """
    This request class is used to submit a request for the latest stock bar data.

    See BaseLatestStockDataRequest for more information on available parameters.
    """

    pass


# ############################## Crypto ################################# #


class BaseLatestCryptoDataRequest(NonEmptyRequest):
    """
    A base request object for retrieving the latest data for crypto. You most likely should not use this directly and
    instead use the asset class specific request objects.

    Attributes:
        symbol_or_symbols (Union[str, List[str]]): The ticker identifier or list of ticker identifiers.
    """

    symbol_or_symbols: Union[str, List[str]]


class LatestCryptoTradeRequest(BaseLatestCryptoDataRequest):
    """
    This request class is used to submit a request for the latest crypto trade data.

    See BaseLatestCryptoDataRequest for more information on available parameters.
    """

    pass


class LatestCryptoQuoteRequest(BaseLatestCryptoDataRequest):
    """
    This request class is used to submit a request for the latest crypto quote data.

    See BaseLatestCryptoDataRequest for more information on available parameters.
    """

    pass


class LatestCryptoBarRequest(BaseLatestCryptoDataRequest):
    """
    This request class is used to submit a request for the latest crypto bar data.

    See BaseLatestCryptoDataRequest for more information on available parameters.
    """

    pass



from datetime import datetime
from typing import Optional, Union, List

from alpaca.common import NonEmptyRequest
from alpaca.data import Adjustment, DataFeed
from alpaca.data.time import TimeFrame

# TODO: Remove duplication?


class BaseBarsRequest(NonEmptyRequest):
    """
    A base request object for retrieving bar data for securities. You most likely should not use this directly and instead
    use the asset class specific request objects.

    Attributes:
        symbol_or_symbols (Union[str, List[str]]): The ticker identifier or list of ticker identifiers.
        start (Optional[datetime]): The beginning of the time interval for desired data.
        end (Optional[datetime]): The end of the time interval for desired data. Defaults to None.
        limit (Optional[int]): Upper limit of number of data points to return. Defaults to None.
        timeframe (TimeFrame): The period over which the bars should be aggregated. (i.e. 5 Min bars, 1 Day bars)
    """

    symbol_or_symbols: Union[str, List[str]]
    start: Optional[Union[datetime, str]]
    end: Optional[Union[datetime, str]]
    limit: Optional[int]
    timeframe: TimeFrame

    # Allows TimeFrame as a non-pydantic BaseModel field
    class Config:
        arbitrary_types_allowed = True


class StockBarsRequest(BaseBarsRequest):
    """
    The request model for retrieving bar data for equities.

    See BaseGetBarsRequest for more information on available parameters.

    Attributes:
        adjustment (Optional[Adjustment]): The type of corporate action data normalization.
        feed (Optional[DataFeed]): The stock data feed to retrieve from.
    """

    adjustment: Optional[Adjustment]
    feed: Optional[DataFeed]


class CryptoBarsRequest(BaseBarsRequest):
    """
    The request model for retrieving bar data for cryptocurrencies.

    See BaseGetBarsRequest for more information on available parameters.
    """

    pass



from enum import Enum


class Exchange(str, Enum):
    """The exchanges that provide data feeds to Alpaca.

    Attributes:
        Z (str): Cboe BZ
        I (str): International Securities Exchange
        M (str): Chicago Stock Exchange
        U (str): Members Exchange
        L (str): Long Term Stock Exchange
        W (str): CBOE
        X (str): NASDAQ OMX PSX
        B (str): NASDAQ OMX BX
        D (str): FINRA ADF
        J (str): Cboe EDGA
        P (str): NYSE Arca
        Q (str): NASDAQ OMX
        S (str): NASDAQ Small Cap
        V (str): IEX
        A (str): NYSE American (AMEX)
        E (str): Market Independent
        N (str): New York Stock Exchange
        T (str): NASDAQ Int
        Y (str): Cboe BYX
        C (str): National Stock Exchange
        H (str): MIAX
        K (str): Cboe EDGX
    """

    Z: str = "Z"
    I: str = "I"
    M: str = "M"
    U: str = "U"
    L: str = "L"
    W: str = "W"
    X: str = "X"
    B: str = "B"
    D: str = "D"
    J: str = "J"
    P: str = "P"
    Q: str = "Q"
    S: str = "S"
    V: str = "V"
    A: str = "A"
    E: str = "E"
    N: str = "N"
    T: str = "T"
    Y: str = "Y"
    C: str = "C"
    H: str = "H"
    K: str = "K"


class DataFeed(str, Enum):
    """Equity market data feeds. OTC and SIP are available with premium data subscriptions.

    Attributes:
        IEX (str): Investor's exchange data feed
        SIP (str): Securities Information Processor feed
        OTC (str): Over the counter feed
    """

    IEX: str = "iex"
    SIP: str = "sip"
    OTC: str = "otc"


class Adjustment(str, Enum):
    """Data normalization based on types of corporate actions.

    Attributes:
        RAW (str): Unadjusted data
        SPLIT (str): Stock-split adjusted data
        DIVIDEND (str): Dividend adjusted data
        ALL (str): Data adjusted for all corporate actions
    """

    RAW: str = "raw"
    SPLIT: str = "split"
    DIVIDEND: str = "dividend"
    ALL: str = "all"



from datetime import datetime
from .models import ValidateBaseModel as BaseModel


class Clock(BaseModel):
    """
    The market clock for US equity markets. Timestamps are in eastern time.

    Attributes:
        timestamp (datetime): The current timestamp.
        is_open (bool): Whether the market is currently open.
        next_open (datetime): The timestamp when the market will next open.
        next_close (datetime): The timestamp when the market will next close.
    """

    timestamp: datetime
    is_open: bool
    next_open: datetime
    next_close: datetime



from datetime import datetime, date
from typing import Any
from .models import ValidateBaseModel as BaseModel


class Calendar(BaseModel):
    """
    The market calendar for equity markets. Describes the market open and close time on a given day.
    """

    date: date
    open: datetime
    close: datetime

    def __init__(self, **data: Any) -> None:
        """
            Converts open and close time strings from %H:%M to a datetime
        Args:
            **data: The raw calendar data from API
        """
        if "date" in data and "open" in data:
            start_datetime_str = data["date"] + " " + data["open"]
            data["open"] = datetime.strptime(start_datetime_str, "%Y-%m-%d %H:%M")

        if "date" in data and "close" in data:
            start_datetime_str = data["date"] + " " + data["close"]
            data["close"] = datetime.strptime(start_datetime_str, "%Y-%m-%d %H:%M")

        super().__init__(**data)



from datetime import datetime
from typing import Union, List, Optional

from alpaca.common import NonEmptyRequest
from alpaca.data import DataFeed


class BaseTradesRequest(NonEmptyRequest):
    """
    A base request object for retrieving trade data for securities. You most likely should not use this directly and instead
    use the asset class specific request objects.

    Attributes:
        symbol_or_symbols (Union[str, List[str]]): The ticker identifier or list of ticker identifiers.
        start (Optional[datetime]): The beginning of the time interval for desired data.
        end (Optional[datetime]): The end of the time interval for desired data.
        limit (Optional[int]): Upper limit of number of data points to return.
    """

    symbol_or_symbols: Union[str, List[str]]
    start: Optional[Union[datetime, str]]
    end: Optional[Union[datetime, str]]
    limit: Optional[int]


class StockTradesRequest(BaseTradesRequest):
    """
    This request class is used to submit a request for stock trade data.

    See BaseTradesRequest for more information on available parameters.

    Attributes:
        feed (Optional[DataFeed]): The stock data feed to retrieve from.
    """

    feed: Optional[DataFeed]


class CryptoTradesRequest(BaseTradesRequest):
    """
    This request class is used to submit a request for crypto trade data.

    See BaseTradesRequest for more information on available parameters.
    """

    pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['alpaca',
 'alpaca.broker',
 'alpaca.broker.models',
 'alpaca.common',
 'alpaca.common.models',
 'alpaca.data',
 'alpaca.data.historical',
 'alpaca.data.live',
 'alpaca.data.models',
 'alpaca.data.requests',
 'alpaca.trading']

package_data = \
{'': ['*']}

install_requires = \
['msgpack>=1.0.3,<2.0.0',
 'pandas>=1.4.1,<2.0.0',
 'pydantic>=1.9.0,<2.0.0',
 'requests>=2.27.1,<3.0.0',
 'websockets>=10.2,<11.0']

setup_kwargs = {
    'name': 'alpaca-py',
    'version': '0.3.0',
    'description': 'The new official Python SDK for the https://alpaca.markets/ API',
    'long_description': "# Alpaca-py\n\n[![Downloads](https://pepy.tech/badge/alpaca-py/month)](https://pepy.tech/project/alpaca-py)\n[![Python Versions](https://img.shields.io/pypi/pyversions/alpaca-py.svg?logo=python&logoColor=white)](https://pypi.org/project/alpaca-py)\n[![GitHub](https://img.shields.io/github/license/alpacahq/alpaca-py?color=blue)](https://github.com/alpacahq/alpaca-py/blob/master/LICENSE.md)\n[![PyPI](https://img.shields.io/pypi/v/alpaca-py?color=blue)](https://pypi.org/project/alpaca-py/)\n### About\n\nAlpaca-py provides an interface for interacting with the various REST and WebSocket endpoints Alpaca offers.\nYou can access both historical and live market data for equities and cryptocurrencies via the Market Data API. \nYou can place trades for both crypto and equities through a uniform interface. Alpaca-py also offers the ability\nto manage your Broker API account by creating accounts, managing funds, and more. \n\nLearn more about the API products [Alpaca]((https://alpaca.markets/)) offers.\n\n**Note: AlpacaPy is in the very early stages of alpha development and is not production ready. Currently AlpacaPy\ninterfaces with only the Market Data API, however the other APIs are coming soon.**\n\n### Installation\n\nAlpaca-py is supported on Python 3.8+.  You can install Alpaca-py using pip.\n\nRun the following command in your terminal.\n\n```shell\n  pip install alpaca-py\n```\n\n\n### Dev setup\n\nThis project is managed via poetry so setup should be just running `poetry install`.\n\nThis repo is using [`pre-commit`](https://pre-commit.com/) to setup some checks to happen at commit time to keep the\nrepo clean. To set these up after you've run `poetry install` just run `poetry run pre-commit install` to have\npre-commit setup these hooks\n\n\n\n",
    'author': 'Rahul Chowdhury',
    'author_email': 'rahul.chowdhury@alpaca.markets',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/alpacahq/alpaca-py',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8.0,<4.0.0',
}


setup(**setup_kwargs)
